#ifndef __Coff_H__
#define __Coff_H__

typedef unsigned int UInt32;
typedef unsigned short UInt16;
typedef unsigned char UInt8;

typedef int Int32;
typedef short Int16;
typedef char Int8;

typedef int Int;
typedef char Char;
typedef void * Ptr;

#define TRUE	(1)
#define FALSE	(0)

#define     SWAP_LOCATION				20u/* Location in COFF file where swap information is kept. */
#define     SECT_DSECT					0x0001u/* dummy section. */
#define     SECT_NOLOAD					0x0002u/* no_load section. */
#define     SECT_BSS					0x0080u/* BSS section. */
#define     SECT_COPY					0x0010u/* COPY section. */

#define     COFF_VERSION				0x00C2u/* Offset in file header where number of sections is present. */
#define     SIZE_OPT_HDR_LOC			16u/* Location in file header for number of bytes in optional header. */
#define     COFF_NAME_LEN				8u/* Length of name. */
#define     SYMTAB_OFFSET				8u/* Offset in file header where symbol table details are present. */
#define     NUM_SECT_OFFSET				2u/* Offset in file header where number of sections is present. */

#define     SIZE_COFF_FILE_HEADER		22u
#define     SIZE_COFF_SYMBOL_ENTRY		18u
#define     SIZE_COFF_SECTION_HEADER	48u

#define     COFF_MAGIC_M3				0x0097u
#define     COFF_MAGIC_64x				0x0099u


#define BYTESWAP_WORD(x) (UInt16) (( (UInt16) ((((UInt16)(x)) << 8u) & 0xFF00u) \
                                   | (UInt16)((((UInt16)(x)) >> 8u) & 0x00FFu)))

#define BYTESWAP_LONG(x) (   (((x) << 24u) & 0xFF000000u)    \
                          |  (((x) <<  8u) & 0x00FF0000u)   \
                          |  (((x) >>  8u) & 0x0000FF00u)   \
                          |  (((x) >> 24u) & 0x000000FFu))

#define WORDSWAP_LONG(x) (   (((x) << 16u) & 0xFFFF0000u)    \
                          |  (((x) >> 16u) & 0x0000FFFFu))

#define SWAP_LONG(value, swap) (((swap) == FALSE) ? (value) : WORDSWAP_LONG (value))

/*!
 *  @brief   Enumerates the different types of COFF loaders supported
 */
typedef enum {
    Coff_LoaderType_File       = 0u,
    /* !< File based COFF loader */
    Coff_LoaderType_Shm        = 1u,
    /* !< COFF loader optimized for shared memory devices */
    Coff_LoaderType_Mem        = 2u,
    /* !< COFF loader that works off a memory buffer instead of file accesses */
    Coff_LoaderType_EndValue   = 3u
    /*!< End delimiter indicating start of invalid values for this enum */
} Coff_LoaderType;

typedef enum {
    Processor_ProcArch_Unknown       = 0u,
    /* !< Indicates that the architecture is not supported. */
    Processor_ProcArch_C55x          = 1u,
    /* !< Indicates that the architecture is C55x. */
    Processor_ProcArch_C64x          = 2u,
    /* !< Indicates that the architecture is C64x. */
    Processor_ProcArch_M3            = 3u,
    /* !< Indicates that the architecture is CORTEX M3. */
    Processor_ProcArch_EndValue      = 4u
    /*!< End delimiter indicating start of invalid values for this enum */
} Processor_ProcArch;


typedef enum {
    AddrType_MasterKnlVirt = 0u,
    /*!< Kernel Virtual address on master processor */
    AddrType_MasterUsrVirt = 1u,
    /*!< User Virtual address on master processor */
    AddrType_MasterPhys    = 2u,
    /*!< Physical address on master processor */
    AddrType_SlaveVirt     = 3u,
    /*!< Virtual address on slave processor */
    AddrType_SlavePhys     = 4u,
    /*!< Physical address on slave processor */
    AddrType_EndValue      = 5u
    /*!< End delimiter indicating start of invalid values for this enum */
} AddrType;

typedef struct ProcMgr_AddrInfo_tag {
    UInt32    addr[AddrType_EndValue];
    UInt32    size;
    int      isCached;
} AddrInfo;

/*!
 *  @brief  File header for a COFF file.
 */
typedef struct Coff_FileHdr_tag {
    UInt16  version;
    /*!< Version ID. indicates the version of the COFF file structure. */
    UInt16  numSections;
    /*!< Number of section headers */
    Int32   dateTime;
    /*!< Time and date stamp. indicates when the file was created. */
    Int32   fpSymTab;
    /*!< Symbol table's starting location in file. */
    Int32   numSymTabEntries;
    /*!< Number of entries in the symbol table. */
    UInt16  numBytesOptHeader;
    /*!< Number of bytes in the optional header. This field is
         either 0 or 28. If it is 0, there is no optional file header. */
    UInt16  flags;
    /*!< Flags (see the File Header Flags table). */
    UInt16  targetId;
    /*!< Target ID. magic number indicates the file can be executed
         in a particular system. This field is checked for validating
         the support of supplied file. */
} Coff_FileHdr;

/*!
 *  @brief  Optional header for coff file format.
 */
typedef struct Coff_OptHdr_tag {
    Int16 magic;
    /*!< Optional file header magic number */
    Int16 version;
    /*!< Version stamp. */
    Int32 sizeExeCode;
    /*!< Size (in bytes) of executable code. */
    Int32 sizeInitData;
    /*!< Size (in bytes) of initialized data. */
    Int32 sizeUninitData;
    /*!< Size (in bytes) of uninitialized data. */
    Int32 entry;
    /*!< Entry point. */
    Int32 addrExe;
    /*!< Beginning address of executable code. */
    Int32 addrInitData;
    /*!< Beginning address of initialized data. */
} Coff_OptHdr;

/*!
 *  @brief  Section header for COFF file format.
 */
typedef struct Coff_SectHdr_tag {
    char   name [COFF_NAME_LEN];
    /*!< This field contains one of the following:
         1) An 8-character section name, padded with nulls, or
         2) A pointer into the string table if the section name is longer than
         8 characters.
         In the latter case the first four bytes of the field are 0. */
    Int32   physicalAddress;
    /*!< Section's physical address. */
    Int32   virtualAddress;
    /*!< Section's virtual address. */
    Int32   size;
    /*!< Section's size in bytes. */
    Int32   fpRawData;
    /*!< File pointer to raw data. */
    Int32   fpReloc;
    /*!< File pointer to relocation entries. */
    Int32   fpLineNum;
    /*!< File pointer to line-number entries. */
    UInt32  numReloc;
    /*!< Number of relocation entries. */
    UInt32  numLine;
    /*!< Number of line-number entries. */
    UInt32  flags;
    /*!< Flags (see the Section Header Flags table) */
    UInt16  reserved;
    /*!< Reserved. */
    UInt16  memPageNum;
    /*!< Memory page number. */
    int    isLoadSection;
    /*!< Flag to indicate that the section is loadable. */
    char *data;
    /*!< Buffer to hold data. */
} Coff_SectHdr;

/*!
 *  @brief  Defines the structure for a symbol table entry.
 */
typedef struct Coff_SymEntry_tag {
    char   name [COFF_NAME_LEN];
    /*!< This field contains one of the following:
         1) An 8-character symbol name, padded with nulls, or
         2) A pointer into the string table if the symbol name is longer than
         8 characters.
         In the latter case the first four bytes of the field are 0. */
    Int32   value;
    /*!< Symbol value; storage class dependent. */
    Int16   sectNum;
    /*!< Section number of the symbol. */
    UInt16  type;
    /*!< Basic and derived type specification. */
    char   storage;
    /*!< Storage class of the symbol. */
    char   numAuxEnt;
    /*!< Number of auxiliary entries (always 0 or 1). If this is '1' then this
         structure is followed by the Auxilliary entry structure (which is of
         the same size as this structure). */
} Coff_SymEntry;

#endif /* __Coff_H__ */

