/**	@file spi_flash_probe.c
 *	@note HangZhou Hikvision System Technology Co., Ltd. All Right Reserved.
 *	@brief ʵspi flash̽
 *
 *	@author		zhaoweibo
 *	@date		2011/06/03
 *
 *
 */
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>
#include <linux/version.h>

#include "hik_spi_flash_m25p16.h"
#include "hik_spi_flash_probe.h"

#include <linux/fs.h>
#include <asm/uaccess.h>
#include <linux/syscalls.h>


#define MTD_FLASH_MAP_DEBUG
#ifdef MTD_FLASH_MAP_DEBUG
#define DB(x)	x
#else
#define DB(x)
#endif

#define MTD_MAX_FLASH_NUMBER	4
#define MTD_DUMMY_BANK_WIDTH	1

struct maps_init_info
{
	struct map_info mapInfo;
	char ** mtdDrv;
	struct mtd_info * mtdInfo;
	char name[32];
};	

static struct maps_init_info maps[MTD_MAX_FLASH_NUMBER];
static unsigned int mapsNum = 0;


static char * sflashDev = "spi_flash";
static char * sflashMtdList[] = {"spi_flash", NULL};


struct mtd_partition spi_mtd_parts[] = 
{
	{
	.name = "sys_params",
	.offset = 0x1F0000,	//2M-64k
	.size   =  0x10000 ,	//64k
	.mask_flags = MTD_WRITEABLE, /* force read-only */	
	},
	
};
static int                   mtd_parts_nb = 1;

/*@fn		fill_flash_info  
 * @brief		fill the basic infomation of spi flash  to maps struct
 * @param[in]  
 *	
 * @param[out] 
 *	expectedDevs	--	the number of device
 * @return	  
 */
static int fill_flash_info(void)
{
	int expectedDevs = 0;
	int devs = 1;
	int i;
	
	/* clear the whole array */
	memset((void*)maps, 0x0, sizeof(maps));
	
	/* gather the SPI flash devices information */
	
	for(i=0; i<devs; i++)
	{
		if (expectedDevs >= MTD_MAX_FLASH_NUMBER)
		{
			printk(KERN_NOTICE "\nERROR: %s - Exceeded MAX MTD flash devices number", __FUNCTION__);		
			break;	
		}
		maps[expectedDevs].mtdDrv = sflashMtdList;	
		maps[expectedDevs].mapInfo.name = sflashDev;
		maps[expectedDevs].mapInfo.phys =  M25P16_BASE_ADDR;
		maps[expectedDevs].mapInfo.size  = SF_M25P16_MAX_SIZE;
		maps[expectedDevs].mapInfo.bankwidth = MTD_DUMMY_BANK_WIDTH;
		
		if ((maps[expectedDevs].mapInfo.phys != 0xFFFFFFFF) && 
		    (maps[expectedDevs].mapInfo.size != 0xFFFFFFFF))
		{
			DB(printk("\nINFO: Found %s %d - base 0x%08x, size 0x%x", maps[expectedDevs].mapInfo.name, i,
   	          (unsigned int)maps[expectedDevs].mapInfo.phys, (unsigned int)maps[expectedDevs].mapInfo.size));
   			++expectedDevs;
		}
		else
		{
			printk(KERN_NOTICE "\nERROR: %s - Failed to get Device Base address and Size (%s %d)", __FUNCTION__, maps[expectedDevs].mapInfo.name, i);
		}
	}
	
	DB(printk("\nINFO: %s - Found %d Flash Devices", __FUNCTION__, expectedDevs));
	return expectedDevs;
}
/*@fn		flash_probe  
 * @brief		probe spi flash
 * @param[in]  
 *	mtdDrv		-- name of spi flash
 *  map_info	-- map struct about spi flash
 *  mtd				-- the pointer of mtd info
 * @param[out] 
 *	int	--	the result 
 * @return	  
 */
static int flash_probe(char ** mtdDrv, struct map_info * map, struct mtd_info ** mtd)
{
	if ((mtdDrv == NULL) || (map == NULL) || (mtd == NULL))
	{
		printk(KERN_NOTICE "\nERROR: NULL pointer parameter at %s entry", __FUNCTION__);
		return -EINVAL;	
	}
	
	/* remap the physical address to a virtual address */
	map->virt = ioremap(map->phys, map->size);	
	if (!map->virt) 
	{
		printk(KERN_NOTICE "\nERROR: Failed to ioremap Flash device at physical base 0x%x.", (unsigned int)map->phys);
		return -EIO;
	}
	
	DB(printk("\nINFO: Io remapped successfully - phy addr = 0x%08x, virt addr = 0x%08x", (unsigned int)map->phys, (unsigned int)map->virt));

	simple_map_init(map);

	*mtd = NULL;
	for(; (!(*mtd) && *mtdDrv); mtdDrv++) 
	{
		DB(printk("\nINFO: Using %s to probe %s at address 0x%08x, size 0x%x, width %d\n", 
                  *mtdDrv, map->name, (unsigned int)map->phys, (unsigned int)map->size, map->bankwidth));
		if ((*mtd = do_map_probe(*mtdDrv, map)))
		{
			DB(printk(" ---  --- detected OK \n"));
			(*mtd)->owner = THIS_MODULE;


			printk("\nparse_mtd_partitions return : %d ", mtd_parts_nb);
			{
				add_mtd_partitions (*mtd, spi_mtd_parts, mtd_parts_nb);
				DB(printk("\nINFO: add mtd partition succeed!"));
				return 0;
			}
		
			if (add_mtd_device(*mtd) != 0)
			{
				printk(KERN_NOTICE "\nERROR: %s - Failed to add the mtd device", __FUNCTION__);
				iounmap((void *)map->virt);
				map->virt = 0;
				return -ENXIO;
			}

			return 0;
		} 
		else 
		{
			DB(printk(" ---  --- Not detected \n"));
		}
	}
 
	iounmap((void *)map->virt);
	map->virt = 0;
	return -ENXIO;
}

static int __init spi_flash_map_init(void)
{	
	int i;
			
	mapsNum = fill_flash_info();
	DB(printk("\nINFO: flash_map_init - detected %d devices",  mapsNum));
		
	for (i=0; i<mapsNum; i++)
	{
		DB(printk("MTD: Initialize the %s device at address 0x%08x", maps[i].mapInfo.name, (unsigned int)maps[i].mapInfo.phys));
		if (flash_probe(maps[i].mtdDrv, &maps[i].mapInfo, &maps[i].mtdInfo) == 0)
		{
			DB(printk(" - OK.\n"));
		}
		else
		{
			maps[i].mtdInfo = NULL;
			DB(printk(" - FAILED!\n"));
		}
	}
	return 0;
}

static void __exit spi_flash_map_exit(void)
{
	int i;
	
	for (i=0; i<mapsNum; i++)
	{
		if (maps[i].mtdInfo) 
		{
			del_mtd_device(maps[i].mtdInfo);
			map_destroy(maps[i].mtdInfo);
		}
	
		if (maps[i].mapInfo.virt) 
		{
			iounmap((void *)maps[i].mapInfo.virt);
			maps[i].mapInfo.virt = 0;
		}
	}
}


module_init(spi_flash_map_init);
module_exit(spi_flash_map_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("MTD map driver for Ti816x platforms");

