/*
 * SPI flash interface
 *
 * Copyright (C) 2008 Atmel Corporation
 * Licensed under the GPL-2 or later.
 */

//#include <common.h>
//#include <malloc.h>
#include <linux/slab.h>
#include <linux/kernel.h>
#include "spi.h"
#include "spi_flash.h"

#include "spi_flash_internal.h"

/*@fn		spi_flash_cmd  
 * @brief		write cmd to spi flash  
 * @param[in]  
 *	spi		--	option of bus and cs
 *	cmd		--	command sent to spi flash
 *	len		--	the length of cmd
 * @param[out] 
 *	response	--	value from spi flash 
 * @return	  
 */
int spi_flash_cmd(struct spi_slave *spi, u8 cmd, void *response, size_t len)
{
    unsigned long flags = SPI_XFER_BEGIN;
    int ret;
    
    if (len == 0)
        flags |= SPI_XFER_END;
    
    ret = spi_xfer(spi, 8, &cmd, NULL, flags);
    if (ret) {
        printk("SF: Failed to send command %02x: %d\n", cmd, ret);
        return ret;
    }
    
    if (len) {
        ret = spi_xfer(spi, len * 8, NULL, response, SPI_XFER_END);
        if (ret)
            printk("SF: Failed to read response (%zu bytes): %d\n",
    				len, ret);
    }
    
    return ret;
}

/*@fn		spi_flash_cmd_read  
 * @brief		read data from spi flash  
 * @param[in]  
 *	spi		--	option of bus and cs
 *	cmd		--	command sent to spi flash
 *	cmd_len	--	the length of cmd
 *	data		--	pointer to buffer that receives data
 *	data_len	--	the length of data buffer
 * @param[out] 
 * @return
 *	0	--	OK	
 *	other--	FAIL
 */
int spi_flash_cmd_read(struct spi_slave *spi, const u8 *cmd,
		size_t cmd_len, void *data, size_t data_len)
{
    unsigned long flags = SPI_XFER_BEGIN;
    int ret;
    
    if (data_len == 0)
        flags |= SPI_XFER_END;
    
    ret = spi_xfer(spi, cmd_len * 8, cmd, NULL, flags);
    if (ret) {
        printk("SF: Failed to send read command (%zu bytes): %d\n",
    			cmd_len, ret);
    } else if (data_len != 0) {
    	ret = spi_xfer(spi, data_len * 8, NULL, data, SPI_XFER_END);
    if (ret)
        printk("SF: Failed to read %zu bytes of data: %d\n",
    				data_len, ret);
    }
    
    return ret;
}

/*@fn		spi_flash_cmd_write  
 * @brief		write data to spi flash  
 * @param[in]  
 *	spi		--	option of bus and cs
 *	cmd		--	command sent to spi flash
 *	cmd_len	--	the length of cmd
 *	data		--	pointer to data which is to be written to flash
 *	data_len	--	the length of data buffer
 * @param[out] 
 * @return
 *	0	--	OK	
 *	other--	FAIL
 */
int spi_flash_cmd_write(struct spi_slave *spi, const u8 *cmd, size_t cmd_len,
		const void *data, size_t data_len)
{
    unsigned long flags = SPI_XFER_BEGIN;
    int ret;
    
    if (data_len == 0)
        flags |= SPI_XFER_END;
    
    ret = spi_xfer(spi, cmd_len * 8, cmd, NULL, flags);
    if (ret) {
        printk("SF: Failed to send read command (%zu bytes): %d\n",
    			cmd_len, ret);
    } else if (data_len != 0) {
    	ret = spi_xfer(spi, data_len * 8, data, NULL, SPI_XFER_END);
    if (ret)
        printk("SF: Failed to read %zu bytes of data: %d\n",
    				data_len, ret);
    }
    
    return ret;
}

/*@fn		spi_flash_read_common  
 * @brief		read common data from spi flash  
 * @param[in]  
 *	spi		--	option of bus and cs
 *	cmd		--	command sent to spi flash
 *	cmd_len	--	the length of cmd
 *	data		--	pointer to buffer that receives data
 *	data_len	--	the length of data buffer
 * @param[out] 
 * @return
 *	0	--	OK	
 *	other--	FAIL
 */
int spi_flash_read_common(struct spi_flash *flash, const u8 *cmd,
		size_t cmd_len, void *data, size_t data_len)
{
    struct spi_slave *spi = flash->spi;
    int ret;
    
    spi_claim_bus(spi);
    ret = spi_flash_cmd_read(spi, cmd, cmd_len, data, data_len);
    spi_release_bus(spi);
    
    return ret;
}

void spi_flash_free(struct spi_flash *flash)
{
    spi_free_slave(flash->spi);
    kfree(flash);
}
